"""
/***************************************************************************
ODTrace
								A QGIS plugin
Trace des flux OD
							-------------------
		begin                : 2012-02-17
		copyright            : (C) 2012 by Gregoire Piffault
		email                : gpiffault@systra.com
***************************************************************************/

/***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************/
"""
from collections import defaultdict
import os.path
# Import the PyQt and QGIS libraries
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from qgis.core import *
# Initialize Qt resources from file resources.py
import resources
# Import the code for the dialog
from odtracedialog import ODTraceDialog

class ODTrace:

	def __init__(self, iface):
		# Save reference to the QGIS interface
		self.iface = iface
		self.conf = {}

	def initGui(self):
		# Create action that will start plugin configuration
		self.action = QAction(QIcon(":/plugins/odtrace/icon.png"), "OD Trace", self.iface.mainWindow())
		# connect the action to the run method
		QObject.connect(self.action, SIGNAL("triggered()"), self.run)

		# Add toolbar button and menu item
		self.iface.addToolBarIcon(self.action)
		self.iface.addPluginToMenu("&OD Trace", self.action)

	def unload(self):
		# Remove the plugin menu item and icon
		self.iface.removePluginMenu("&OD Trace",self.action)
		self.iface.removeToolBarIcon(self.action)

	# run method that performs all the real work
	def run(self):
		# retrieve point layers
		pointlayers = [layer for layer in QgsMapLayerRegistry.instance().mapLayers().itervalues() if layer.type() == QgsMapLayer.VectorLayer and layer.geometryType() == QGis.Point]
		# create and show the dialog
		dlg = ODTraceDialog(pointlayers)
		# show the dialog
		dlg.show()
		result = dlg.exec_()
		# See if OK was pressed
		if result == 1:
			# Read matrix from influx
			influx = open(unicode(dlg.ui.influx.text()))
			headers = influx.next().rstrip("\r\n").split("\t")[2:]
			matrix = defaultdict(lambda:[0] * 2 * len(headers))
			for line in influx:
				line = line.rstrip("\r\n").split("\t")
				o, d = unicode(line[0]), unicode(line[1])
				values = [float(v) for v in line[2:]]
				if o < d:
					matrix[(o, d)][:len(headers)] = values
				else:
					matrix[(d, o)][len(headers):] = values
			# Map identifier to points
			selected_field = dlg.selected_field
			provider = dlg.selected_pointlayer.dataProvider()
			provider.select([selected_field])
			points = dict([(unicode(feature.attributeMap()[selected_field].toString()), feature.geometry().asPoint()) for feature in provider])
			# Retrieve selected points
			selected_points = [unicode(feature.attributeMap()[selected_field].toString()) for feature in dlg.selected_pointlayer.selectedFeatures()]
			if not selected_points:
				selected_points = points.keys()
			# Open outshape in read mode if exists and map od to polyline
			existing_features = {}
			if os.path.isfile(dlg.outshape):
				# Look for already opened outshape
				opened_layers = [layer for layer in QgsMapLayerRegistry.instance().mapLayers().values() if layer.source() == dlg.outshape]
				if opened_layers:
					layer = opened_layers[0]
				# Otherwise open it
				else:
					layer = QgsVectorLayer(dlg.outshape, os.path.basename(unicode(dlg.outshape)).rstrip(".shp"), "ogr")
					QgsMapLayerRegistry.instance().addMapLayer(layer)
				provider = layer.dataProvider()
				provider.select(provider.fields().keys())
				# Map od to polylines
				for feature in provider:
					o = unicode(feature.attributeMap()[0].toString()).strip()
					d = unicode(feature.attributeMap()[1].toString()).strip()
					existing_features[(o, d)] = feature.geometry().asPolyline()
				# Close every instance of outshape
				for opened_id in [id for id, layer in QgsMapLayerRegistry.instance().mapLayers().items() if layer.source() == dlg.outshape]:
					QgsMapLayerRegistry.instance().removeMapLayer(opened_id)
			# Define fields
			if not dlg.ui.sum_forw_back.isChecked():
				headers = headers + [h + "_r" for h in headers]
			fields = {0:QgsField("O", QVariant.String), 1:QgsField("D", QVariant.String)}
			fields.update(dict([(i + 2, QgsField(h, QVariant.Double)) for i, h in enumerate(headers)]))
			# Open outshape in write mode
			writer = QgsVectorFileWriter(dlg.outshape, 
										dlg.encoding, 
										fields, 
										QGis.WKBLineString, dlg.selected_pointlayer.crs(), 
										"ESRI Shapefile")
			# QMessageBox.warning(self.iface.mainWindow(), "OD Trace", repr(writer.errorMessage()))
			# Process ODs
			for (o, d), values in matrix.iteritems():
				# Skip links below threshold
				if not check_threshold(values, dlg.influx_threshold):
					continue
				# Skip links with no selected points
				if not (o in selected_points or d in selected_points):
					continue
				# Skip o == d
				if o == d:
					continue
				# Skip unknown points
				try:
					po, pd = points[o], points[d]
				except:
					continue
				feature = QgsFeature()
				# Check if line alredy exists
				if (o, d) in existing_features:
					feature.setGeometry(QgsGeometry.fromPolyline(existing_features[(o, d)]))
				else:
					feature.setGeometry(QgsGeometry.fromPolyline([po, pd]))
				# Sum forward and backward fields if necessary
				if dlg.ui.sum_forw_back.isChecked():
					half_len = len(values) // 2
					values = [values[i] + values[i+half_len] for i in range(half_len)]
				# Populate attributes
				for i, v in enumerate([o, d] + values):
					feature.addAttribute(i, QVariant(v))
				# Write feature
				writer.addFeature(feature)
				# QMessageBox.warning(self.iface.mainWindow(), "OD Trace", "here man")
			del writer
			# Add layer to canvas
			QgsMapLayerRegistry.instance().addMapLayer(QgsVectorLayer(dlg.outshape, os.path.basename(unicode(dlg.outshape)).rstrip(".shp"), "ogr"))

def check_threshold(values, threshold):
	for value in values:
		if value >= threshold:
			return True
	return False
